"""
Manage FreeBSD services.
"""

from __future__ import annotations

from typing_extensions import List, Optional, Union

from pyinfra import host
from pyinfra.api import QuoteString, StringCommand, operation
from pyinfra.api.exceptions import OperationValueError
from pyinfra.facts.freebsd import ServiceScript, ServiceStatus

SRV_STARTED: str = "started"
SRV_STOPPED: str = "stopped"
SRV_RESTARTED: str = "restarted"
SRV_RELOADED: str = "reloaded"
SRV_CUSTOM: str = "custom"


@operation()
def service(
    srvname: str,
    jail: Optional[str] = None,
    srvstate: str = SRV_STARTED,
    command: Optional[Union[str, List[str]]] = None,
    environment: Optional[List[str]] = None,
    verbose: bool = False,
):
    """
    Control (start/stop/etc.) ``rc(8)`` scripts.

    + srvname: Service.
    + jail: See ``-j`` in ``service(8)``.
    + srvstate: Desire state of the service.
    + command: When ``srvstate`` is ``custom``, the command to execute.
    + environment: See ``-E`` in ``service(8)``.
    + verbose: See ``-v`` in ``service(8)``.

    States:
        There are a few states you can use to manipulate the service:

        - started: The service must be started.
        - stopped: The service must be stopped.
        - restarted: The service must be restarted.
        - reloaded: The service must be reloaded.
        - custom: Run a custom command for this service.

    **Examples:**

    .. code:: python

        # Start a service.
        service.service(
            "beanstalkd",
            srvstate="started"
        )

        # Execute a custom command.
        service.service(
            "sopel",
            srvstate="custom",
            command="configure"
        )
    """

    if not host.get_fact(ServiceScript, srvname=srvname, jail=jail):
        host.noop(f"Cannot find rc(8) script '{srvname}'")
        return

    args: List[Union[str, "QuoteString"]] = []

    args.append("service")

    if verbose:
        args.append("-v")

    if jail is not None:
        args.extend(["-j", QuoteString(jail)])

    if environment is not None:
        for env_var in environment:
            args.extend(["-E", QuoteString(env_var)])

    if srvstate == SRV_STARTED:
        if host.get_fact(ServiceStatus, srvname=srvname, jail=jail):
            host.noop(f"Service '{srvname}' already started")
            return

        args.extend([QuoteString(srvname), "start"])

    elif srvstate == SRV_STOPPED:
        if not host.get_fact(ServiceStatus, srvname=srvname, jail=jail):
            host.noop(f"Service '{srvname}' already stopped")
            return

        args.extend([QuoteString(srvname), "stop"])
    elif srvstate == SRV_RESTARTED:
        args.extend([QuoteString(srvname), "restart"])

    elif srvstate == SRV_RELOADED:
        args.extend([QuoteString(srvname), "reload"])

    elif srvstate == SRV_CUSTOM:
        args.append(QuoteString(srvname))

        if command is not None:
            if isinstance(command, str):
                command = [command]

            args.extend([QuoteString(c) for c in command])

    else:
        raise OperationValueError(f"Invalid service command: {srvstate}")

    yield StringCommand(*args)



from __future__ import annotations

from .postgres import PostgresDatabases, PostgresRoles


class PostgresqlRoles(PostgresRoles):
    deprecated = True


class PostgresqlDatabases(PostgresDatabases):
    deprecated = True



from __future__ import annotations

import json
import os
import re
import shutil
from datetime import datetime
from tempfile import mkdtemp
from typing import Dict, Iterable, List, Optional, Tuple, Union

from dateutil.parser import parse as parse_date
from distro import distro
from typing_extensions import TypedDict, override

from pyinfra import host
from pyinfra.api import FactBase, ShortFactBase
from pyinfra.api.util import try_int
from pyinfra.facts import crontab

ISO_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S%z"


class User(FactBase):
    """
    Returns the name of the current user.
    """

    @override
    def command(self):
        return "echo $USER"


class Home(FactBase[Optional[str]]):
    """
    Returns the home directory of the given user, or the current user if no user is given.
    """

    @override
    def command(self, user=""):
        return f"echo ~{user}"


class Path(FactBase):
    """
    Returns the path environment variable of the current user.
    """

    @override
    def command(self):
        return "echo $PATH"


class TmpDir(FactBase):
    """
    Returns the temporary directory of the current server.

    According to POSIX standards, checks environment variables in this order:
    1. TMPDIR (if set and accessible)
    2. TMP (if set and accessible)
    3. TEMP (if set and accessible)
    4. Falls back to empty string
    """

    @override
    def command(self):
        return """
if [ -n "$TMPDIR" ] && [ -d "$TMPDIR" ] && [ -w "$TMPDIR" ]; then
    echo "$TMPDIR"
elif [ -n "$TMP" ] && [ -d "$TMP" ] && [ -w "$TMP" ]; then
    echo "$TMP"
elif [ -n "$TEMP" ] && [ -d "$TEMP" ] && [ -w "$TEMP" ]; then
    echo "$TEMP"
else
    echo ""
fi
        """.strip()


class Hostname(FactBase):
    """
    Returns the current hostname of the server.
    """

    @override
    def command(self):
        return "uname -n"


class Kernel(FactBase):
    """
    Returns the kernel name according to ``uname``.
    """

    @override
    def command(self):
        return "uname -s"


class KernelVersion(FactBase):
    """
    Returns the kernel version according to ``uname``.
    """

    @override
    def command(self):
        return "uname -r"


# Deprecated/renamed -> Kernel
class Os(FactBase[str]):
    """
    Returns the OS name according to ``uname``.

    .. warning::
        This fact is deprecated/renamed, please use the ``server.Kernel`` fact.
    """

    @override
    def command(self):
        return "uname -s"


# Deprecated/renamed -> KernelVersion
class OsVersion(FactBase[str]):
    """
    Returns the OS version according to ``uname``.

    .. warning::
        This fact is deprecated/renamed, please use the ``server.KernelVersion`` fact.
    """

    @override
    def command(self):
        return "uname -r"


class Arch(FactBase[str]):
    """
    Returns the system architecture according to ``uname``.
    """

    # ``uname -p`` is not portable and returns ``unknown`` on Debian.
    # ``uname -m`` works on most Linux and BSD systems.
    @override
    def command(self):
        return "uname -m"


class Command(FactBase[str]):
    """
    Returns the raw output lines of a given command.
    """

    @override
    def command(self, command):
        return command


class Which(FactBase[Optional[str]]):
    """
    Returns the path of a given command according to `command -v`, if available.
    """

    @override
    def command(self, command):
        return "command -v {0} || true".format(command)


class Date(FactBase[datetime]):
    """
    Returns the current datetime on the server.
    """

    default = datetime.now

    @override
    def command(self):
        return f"date +'{ISO_DATE_FORMAT}'"

    @override
    def process(self, output) -> datetime:
        return datetime.strptime(list(output)[0], ISO_DATE_FORMAT)


class MacosVersion(FactBase[str]):
    """
    Returns the installed MacOS version.
    """

    @override
    def requires_command(self) -> str:
        return "sw_vers"

    @override
    def command(self):
        return "sw_vers -productVersion"


class MountsDict(TypedDict):
    device: str
    type: str
    options: list[str]


class Mounts(FactBase[Dict[str, MountsDict]]):
    """
    Returns a dictionary of mounted filesystems and information.

    .. code:: python

        {
            "/": {
                "device": "/dev/mv2",
                "type": "ext4",
                "options": [
                    "rw",
                    "relatime"
                ]
            },
        }
    """

    default = dict

    @override
    def command(self) -> str:
        self._kernel = host.get_fact(Kernel)

        if self._kernel.strip() == "FreeBSD":
            return "mount -p --libxo json"
        else:
            return "cat /proc/self/mountinfo"

    @override
    def process(self, output) -> dict[str, MountsDict]:
        devices: dict[str, MountsDict] = {}

        def unescape_octal(match: re.Match) -> str:
            s = match.group(0)[1:]  # skip the backslash
            return chr(int(s, base=8))

        def replace_octal(s: str) -> str:
            """
            Unescape strings encoded by linux's string_escape_mem with ESCAPE_OCTAL flag.
            """
            return re.sub(r"\\[0-7]{3}", unescape_octal, s)

        if self._kernel == "FreeBSD":
            full_output = "\n".join(output)
            json_output = json.loads(full_output)
            mount_fstab = json_output["mount"]["fstab"]

            for entry in mount_fstab:
                path = entry["mntpoint"]
                type_ = entry["fstype"]
                device = entry["device"]
                options = [option.strip() for option in entry["opts"].split(",")]

                devices[path] = {"device": device, "type": type_, "options": options}

            return devices

        for line in output:
            # ignore mount ID, parent ID, major:minor, root
            _, _, _, _, mount_point, mount_options, line = line.split(sep=" ", maxsplit=6)

            # ignore optional tags "shared", "master", "propagate_from" and "unbindable"
            while True:
                optional, line = line.split(sep=" ", maxsplit=1)
                if optional == "-":
                    break

            fs_type, mount_source, super_options = line.split(sep=" ")

            mount_options = mount_options.split(sep=",")

            # escaped: mount_point, mount_source, super_options
            # these strings can contain characters encoded in octal, e.g. '\054' for ','
            mount_point = replace_octal(mount_point)
            mount_source = replace_octal(mount_source)

            # mount_options will override ro/rw and can be different than the super block options
            # filter them, so they don't appear twice
            super_options = [
                replace_octal(opt)
                for opt in super_options.split(sep=",")
                if opt not in ["ro", "rw"]
            ]

            devices[mount_point] = {
                "device": mount_source,
                "type": fs_type,
                "options": mount_options + super_options,
            }

        return devices


class Port(FactBase[Union[Tuple[str, int], Tuple[None, None]]]):
    """
    Returns the process occuping a port and its PID
    """

    @override
    def command(self, port: int) -> str:
        return f"ss -lptnH 'src :{port}'"

    @override
    def process(self, output: Iterable[str]) -> Union[Tuple[str, int], Tuple[None, None]]:
        for line in output:
            proc, pid = line.split('"')[1], int(line.split("pid=")[1].split(",")[0])
            return (proc, pid)
        return None, None


class KernelModules(FactBase):
    """
    Returns a dictionary of kernel module name -> info.

    .. code:: python

        {
            "module_name": {
                "size": 0,
                "instances": 0,
                "state": "Live",
            },
        }
    """

    @override
    def command(self):
        return "! test -f /proc/modules || cat /proc/modules"

    default = dict

    @override
    def process(self, output):
        modules = {}

        for line in output:
            name, size, instances, depends, state, _ = line.split(" ", 5)
            instances = int(instances)

            module = {
                "size": size,
                "instances": instances,
                "state": state,
            }

            if depends != "-":
                module["depends"] = [value for value in depends.split(",") if value]

            modules[name] = module

        return modules


class LsbRelease(FactBase):
    """
    Returns a dictionary of release information using ``lsb_release``.

    .. code:: python

        {
            "id": "Ubuntu",
            "description": "Ubuntu 18.04.2 LTS",
            "release": "18.04",
            "codename": "bionic",
            ...
        }
    """

    @override
    def command(self):
        return "lsb_release -ca"

    @override
    def requires_command(self):
        return "lsb_release"

    @override
    def process(self, output):
        items = {}

        for line in output:
            if ":" not in line:
                continue

            key, value = line.split(":", 1)

            key = key.strip().lower()

            # Turn "distributor id" into "id"
            if " " in key:
                key = key.split(" ")[-1]

            value = value.strip()

            items[key] = value

        return items


class OsRelease(FactBase):
    """
    Returns a dictionary of release information stored in ``/etc/os-release``.

    .. code:: python

        {
          "name": "EndeavourOS",
          "pretty_name": "EndeavourOS",
          "id": "endeavouros",
          "id_like": "arch",
          "build_id": "2024.06.25",
          ...
        }
    """

    @override
    def command(self):
        return "cat /etc/os-release"

    @override
    def process(self, output):
        items = {}

        for line in output:
            if "=" in line:
                key, value = line.split("=", 1)
                items[key.strip().lower()] = value.strip().strip('"')

        return items


class Sysctl(FactBase):
    """
    Returns a dictionary of sysctl settings and values.

    .. code:: python

        {
            "fs.inotify.max_queued_events": 16384,
            "fs.inode-state": [
                44565,
                360,
            ],
        }
    """

    default = dict

    @override
    def command(self, keys=None):
        if keys is None:
            return "sysctl -a"
        return f"sysctl {' '.join(keys)}"

    @override
    def process(self, output):
        sysctls = {}

        for line in output:
            key = values = None

            if "=" in line:
                key, values = line.split("=", 1)
            elif ":" in line:
                key, values = line.split(":", 1)
            else:
                continue  # pragma: no cover

            if key and values:
                key = key.strip()
                values = values.strip()

                if re.match(r"^[a-zA-Z0-9_\-\.\s]+$", values):
                    values = [try_int(item.strip()) for item in values.split()]

                    if len(values) == 1:
                        values = values[0]

                sysctls[key] = values

        return sysctls


class Groups(FactBase[List[str]]):
    """
    Returns a list of groups on the system.
    """

    @override
    def command(self):
        return "cat /etc/group"

    default = list

    @override
    def process(self, output) -> list[str]:
        groups: list[str] = []

        for line in output:
            if ":" in line:
                groups.append(line.split(":")[0])

        return groups


# for compatibility
CrontabDict = crontab.CrontabDict
Crontab = crontab.Crontab


class Users(FactBase):
    """
    Returns a dictionary of users -> details.

    .. code:: python

        {
            "user_name": {
                "comment": "Full Name",
                "home": "/home/user_name",
                "shell": "/bin/bash,
                "group": "main_user_group",
                "groups": [
                    "other",
                    "groups"
                ],
                "uid": user_id,
                "gid": main_user_group_id,
                "lastlog": last_login_time,
                "password": encrypted_password,
            },
        }
    """

    @override
    def command(self):
        return """

        for i in `cat /etc/passwd | cut -d: -f1`; do
            ENTRY=`grep ^$i: /etc/passwd`;
            LASTLOG=`(((lastlog -u $i || lastlogin $i) 2> /dev/null) | grep ^$i | tr -s ' ')`;
            PASSWORD=`(grep ^$i: /etc/shadow || grep ^$i: /etc/master.passwd) 2> /dev/null | cut -d: -f2`;
            echo "$ENTRY|`id -gn $i`|`id -Gn $i`|$LASTLOG|$PASSWORD";
        done
    """.strip()  # noqa

    default = dict

    @override
    def process(self, output):
        users = {}
        rex = r"[A-Z][a-z]{2} [A-Z][a-z]{2} {1,2}\d+ .+$"

        for line in output:
            entry, group, user_groups, lastlog, password = line.rsplit("|", 4)

            if entry:
                # Parse out the comment/home/shell
                entries = entry.split(":")

                # Parse groups
                groups = []
                for group_name in user_groups.split(" "):
                    # We only want secondary groups here
                    if group_name and group_name != group:
                        groups.append(group_name)

                raw_login_time = None
                login_time = None

                # Parse lastlog info
                # lastlog output varies, which is why I use regex to match login time
                login = re.search(rex, lastlog)
                if login:
                    raw_login_time = login.group()
                    login_time = parse_date(raw_login_time)

                users[entries[0]] = {
                    "home": entries[5] or None,
                    "comment": entries[4] or None,
                    "shell": entries[6] or None,
                    "group": group,
                    "groups": groups,
                    "uid": int(entries[2]),
                    "gid": int(entries[3]),
                    "lastlog": raw_login_time,
                    "login_time": login_time,
                    "password": password,
                }

        return users


class LinuxDistributionDict(TypedDict):
    name: Optional[str]
    major: Optional[int]
    minor: Optional[int]
    release_meta: Dict


class LinuxDistribution(FactBase[LinuxDistributionDict]):
    """
    Returns a dict of the Linux distribution version. Ubuntu, Debian, CentOS,
    Fedora & Gentoo currently. Also contains any key/value items located in
    release files.

    .. code:: python

        {
            "name": "Ubuntu",
            "major": 20,
            "minor": 04,
            "release_meta": {
                "CODENAME": "focal",
                "ID_LIKE": "debian",
                ...
            }
        }
    """

    @override
    def command(self) -> str:
        return (
            "cd /etc/ && for file in $(ls -pdL *-release | grep -v /); "
            'do echo "/etc/${file}"; cat "/etc/${file}"; echo ---; '
            "done"
        )

    name_to_pretty_name = {
        "alpine": "Alpine",
        "centos": "CentOS",
        "fedora": "Fedora",
        "gentoo": "Gentoo",
        "opensuse": "openSUSE",
        "rhel": "RedHat",
        "ubuntu": "Ubuntu",
        "debian": "Debian",
    }

    @override
    @staticmethod
    def default() -> LinuxDistributionDict:
        return {
            "name": None,
            "major": None,
            "minor": None,
            "release_meta": {},
        }

    @override
    def process(self, output) -> LinuxDistributionDict:
        parts = {}
        for part in "\n".join(output).strip().split("---"):
            if not part.strip():
                continue
            try:
                filename, content = part.strip().split("\n", 1)
                parts[filename] = content
            except ValueError:
                # skip empty files
                # for instance arch linux as an empty file at /etc/arch-release
                continue

        release_info = self.default()
        if not parts:
            return release_info

        temp_root = mkdtemp()
        try:
            temp_etc_dir = os.path.join(temp_root, "etc")
            os.mkdir(temp_etc_dir)

            for filename, content in parts.items():
                with open(
                    os.path.join(temp_etc_dir, os.path.basename(filename)),
                    "w",
                    encoding="utf-8",
                ) as fp:
                    fp.write(content)

            parsed = distro.LinuxDistribution(
                root_dir=temp_root,
                include_lsb=False,
                include_uname=False,
            )

            release_meta = {key.upper(): value for key, value in parsed.os_release_info().items()}
            # Distro 1.7+ adds this, breaking tests
            # TODO: fix this!
            release_meta.pop("RELEASE_CODENAME", None)

            release_info.update(
                {
                    "name": self.name_to_pretty_name.get(parsed.id(), parsed.name()),
                    "major": try_int(parsed.major_version()) or None,
                    "minor": try_int(parsed.minor_version()) or None,
                    "release_meta": release_meta,
                },
            )

        finally:
            shutil.rmtree(temp_root)

        return release_info


class LinuxName(ShortFactBase[str]):
    """
    Returns the name of the Linux distribution. Shortcut for
    ``host.get_fact(LinuxDistribution)['name']``.
    """

    fact = LinuxDistribution

    @override
    def process_data(self, data) -> str:
        return data["name"]


class SelinuxDict(TypedDict):
    mode: Optional[str]


class Selinux(FactBase[SelinuxDict]):
    """
    Discovers the SELinux related facts on the target host.

    .. code:: python

        {
            "mode": "enabled",
        }
    """

    @override
    def command(self):
        return "sestatus"

    @override
    def requires_command(self) -> str:
        return "sestatus"

    @override
    @staticmethod
    def default() -> SelinuxDict:
        return {
            "mode": None,
        }

    @override
    def process(self, output) -> SelinuxDict:
        selinux_info = self.default()

        match = re.match(r"^SELinux status:\s+(\S+)", "\n".join(output))

        if not match:
            return selinux_info

        selinux_info["mode"] = match.group(1)

        return selinux_info


class LinuxGui(FactBase[List[str]]):
    """
    Returns a list of available Linux GUIs.
    """

    @override
    def command(self):
        return "ls /usr/bin/*session || true"

    default = list

    known_gui_binaries = {
        "/usr/bin/gnome-session": "GNOME",
        "/usr/bin/mate-session": "MATE",
        "/usr/bin/lxsession": "LXDE",
        "/usr/bin/plasma_session": "KDE Plasma",
        "/usr/bin/xfce4-session": "XFCE 4",
    }

    @override
    def process(self, output) -> list[str]:
        gui_names = []

        for line in output:
            gui_name = self.known_gui_binaries.get(line)
            if gui_name:
                gui_names.append(gui_name)

        return gui_names


class HasGui(ShortFactBase[bool]):
    """
    Returns a boolean indicating the remote side has GUI capabilities. Linux only.
    """

    fact = LinuxGui

    @override
    def process_data(self, data) -> bool:
        return len(data) > 0


class Locales(FactBase[List[str]]):
    """
    Returns installed locales on the target host.

    .. code:: python

        ["C.UTF-8", "en_US.UTF-8"]
    """

    @override
    def command(self) -> str:
        return "locale -a"

    @override
    def requires_command(self) -> str:
        return "locale"

    default = list

    @override
    def process(self, output) -> list[str]:
        # replace utf8 with UTF-8 to match names in /etc/locale.gen
        # return a list of enabled locales
        return [line.replace("utf8", "UTF-8") for line in output]


class SecurityLimits(FactBase):
    """
    Returns a list of security limits on the target host.

    .. code:: python

        [
            {
                "domain": "*",
                "limit_type": "soft",
                "item": "nofile",
                "value": "1048576"
            },
            {
                "domain": "*",
                "limit_type": "hard",
                "item": "nofile",
                "value": "1048576"
            },
            {
                "domain": "root",
                "limit_type": "soft",
                "item": "nofile",
                "value": "1048576"
            },
            {
                "domain": "root",
                "limit_type": "hard",
                "item": "nofile",
                "value": "1048576"
            },
            {
                "domain": "*",
                "limit_type": "soft",
                "item": "memlock",
                "value": "unlimited"
            },
            {
                "domain": "*",
                "limit_type": "hard",
                "item": "memlock",
                "value": "unlimited"
            },
            {
                "domain": "root",
                "limit_type": "soft",
                "item": "memlock",
                "value": "unlimited"
            },
            {
                "domain": "root",
                "limit_type": "hard",
                "item": "memlock",
                "value": "unlimited"
            }
        ]
    """

    @override
    def command(self):
        return "cat /etc/security/limits.conf"

    default = list

    @override
    def process(self, output):
        limits = []

        for line in output:
            if line.startswith("#") or not len(line.strip()):
                continue

            domain, limit_type, item, value = line.split()

            limits.append(
                {
                    "domain": domain,
                    "limit_type": limit_type,
                    "item": item,
                    "value": value,
                },
            )

        return limits


class RebootRequired(FactBase[bool]):
    """
    Returns a boolean indicating whether the system requires a reboot.

    On Linux systems:
    - Checks /var/run/reboot-required and /var/run/reboot-required.pkgs
    - On Alpine Linux, compares installed kernel with running kernel

    On FreeBSD systems:
    - Compares running kernel version with installed kernel version
    """

    @override
    def command(self) -> str:
        return """
# Get OS type
OS_TYPE=$(uname -s)
if [ "$OS_TYPE" = "Linux" ]; then
    # Check if it's Alpine Linux
    if [ -f /etc/alpine-release ]; then
        RUNNING_KERNEL=$(uname -r)
        INSTALLED_KERNEL=$(ls -1 /lib/modules | sort -V | tail -n1)
        if [ "$RUNNING_KERNEL" != "$INSTALLED_KERNEL" ]; then
            echo "reboot_required"
            exit 0
        fi
    else
        # Check standard Linux reboot required files
        if [ -f /var/run/reboot-required ] || [ -f /var/run/reboot-required.pkgs ]; then
            echo "reboot_required"
            exit 0
        fi
    fi
elif [ "$OS_TYPE" = "FreeBSD" ]; then
    RUNNING_VERSION=$(freebsd-version -r)
    INSTALLED_VERSION=$(freebsd-version -k)
    if [ "$RUNNING_VERSION" != "$INSTALLED_VERSION" ]; then
        echo "reboot_required"
        exit 0
    fi
fi
echo "no_reboot_required"
"""

    @override
    def process(self, output) -> bool:
        return list(output)[0].strip() == "reboot_required"



# encoding: utf8
from __future__ import annotations

from typing_extensions import override

from pyinfra.api import FactBase

from .util.packaging import parse_packages

NPM_REGEX = r"^[└├]\─\─\s([a-zA-Z0-9\-]+)@([0-9\.]+)$"


class NpmPackages(FactBase):
    """
    Returns a dict of installed npm packages globally or in a given directory:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    default = dict

    @override
    def requires_command(self, directory=None) -> str:
        return "npm"

    @override
    def command(self, directory=None):
        if directory:
            return ("! test -d {0} || (cd {0} && npm list -g --depth=0)").format(directory)
        return "npm list -g --depth=0"

    @override
    def process(self, output):
        return parse_packages(NPM_REGEX, output)



"""
Manage FreeBSD packages.
"""

from __future__ import annotations

from typing_extensions import List, Optional, Union

from pyinfra import host
from pyinfra.api import QuoteString, StringCommand, operation
from pyinfra.facts.freebsd import PkgPackage


@operation()
def update(jail: Optional[str] = None, force: bool = False, reponame: Optional[str] = None):
    """
    Update the local catalogues of the enabled package repositories.

    + jail: See ``-j`` in ``pkg(8)``.
    + force: See ``-f`` in ``pkg-update(8)``.
    + reponame: See ``-r`` in ``pkg-update(8)``

    **Examples:**

    .. code:: python

        # host
        pkg.update()

        # jail
        pkg.update(
            jail="nginx"
        )
    """

    args: List[Union[str, "QuoteString"]] = []

    args.append("pkg")

    if jail is not None:
        args.extend(["-j", QuoteString(jail)])

    args.extend(["update"])

    if force:
        args.append("-f")

    if reponame is not None:
        args.extend(["-r", QuoteString(reponame)])

    yield StringCommand(*args)


@operation()
def upgrade(jail: Optional[str] = None, force: bool = False, reponame: Optional[str] = None):
    """
    Perform upgrades of package software distributions.

    + jail: See ``-j`` in ``pkg(8)``.
    + force: See ``-f`` in ``pkg-upgrade(8)``.
    + reponame: See ``-r`` in ``pkg-upgrade(8)``.

    **Examples:**

    .. code:: python

        # host
        pkg.upgrade()

        # jail
        pkg.upgrade(
            jail="nginx"
        )
    """

    args: List[Union[str, "QuoteString"]] = []

    args.append("pkg")

    if jail is not None:
        args.extend(["-j", QuoteString(jail)])

    args.extend(["upgrade", "-y"])

    if force:
        args.append("-f")

    if reponame is not None:
        args.extend(["-r", QuoteString(reponame)])

    yield StringCommand(*args)


@operation()
def install(package: str, jail: Optional[str] = None, reponame: Optional[str] = None):
    """
    Install packages from remote packages repositories or local archives.

    + package: Package to install.
    + jail: See ``-j`` in ``pkg(8)``.
    + reponame: See ``-r`` in ``pkg-install(8)``.

    **Example:**

    .. code:: python

        pkg.install("nginx")
    """

    if host.get_fact(PkgPackage, package=package, jail=jail):
        host.noop(f"Package '{package}' already installed")
        return

    args: List[Union[str, "QuoteString"]] = []

    args.append("pkg")

    if jail is not None:
        args.extend(["-j", QuoteString(jail)])

    args.extend(["install", "-y"])

    if reponame is not None:
        args.extend(["-r", QuoteString(reponame)])

    args.extend(["--", QuoteString(package)])

    yield StringCommand(*args)


@operation()
def remove(package: str, jail: Optional[str] = None):
    """
    Deletes packages from the database and the system.

    + package: Package to remove.
    + jail: See ``-j`` in ``pkg(8)``.

    **Example:**

    .. code:: python

        pkg.remove("nginx")
    """

    if not host.get_fact(PkgPackage, package=package, jail=jail):
        host.noop(f"Package '{package}' cannot be found")
        return

    args: List[Union[str, "QuoteString"]] = []

    args.append("pkg")

    if jail is not None:
        args.extend(["-j", QuoteString(jail)])

    args.extend(["remove", "-y"])

    args.extend(["--", QuoteString(package)])

    yield StringCommand(*args)


@operation()
def autoremove(jail: Optional[str] = None):
    """
    Remove orphan packages.

    + jail: See ``-j`` in ``pkg(8)``.

    **Example:**

    .. code:: python

        pkg.autoremove()
    """

    args: List[Union[str, "QuoteString"]] = []

    args.append("pkg")

    if jail is not None:
        args.extend(["-j", QuoteString(jail)])

    args.extend(["autoremove", "-y"])

    yield StringCommand(*args)


@operation()
def clean(all_pkg: bool = False, jail: Optional[str] = None):
    """
    Clean the local cache of fetched remote packages.

    + all_pkg: See ``-a`` in ``pkg-clean(8)``.
    + jail: See ``-j`` in ``pkg(8)``.

    **Example:**

    .. code:: python

        pkg.clean(
            all_pkg=True
        )
    """

    args: List[Union[str, "QuoteString"]] = []

    args.append("pkg")

    if jail is not None:
        args.extend(["-j", QuoteString(jail)])

    args.extend(["clean", "-y"])

    if all_pkg:
        args.append("-a")

    yield StringCommand(*args)



"""
Podman based integration tests.
"""

import pytest


@pytest.mark.end_to_end
@pytest.mark.end_to_end_podman
def test_int_podman_install_package_ubuntu(helpers):
    helpers.run_check_output(
        "pyinfra -y --chdir examples @podman/ubuntu:22.04 apt.packages iftop update=true",
        expected_lines=["podman build complete"],
    )



# from https://stackoverflow.com/a/60708339, but with a few modifications
from __future__ import annotations  # for | in type hints

import re

units = {
    "B": 1,
    "KB": 10**3,
    "MB": 10**6,
    "GB": 10**9,
    "TB": 10**12,
    "KIB": 2**10,
    "MIB": 2**20,
    "GIB": 2**30,
    "TIB": 2**40,
}


def parse_human_readable_size(size: str) -> int:
    size = size.upper()
    if not re.match(r" ", size):
        size = re.sub(r"([KMGT]?I?[B])", r" \1", size)
    number, unit = [string.strip() for string in size.split()]
    return int(float(number) * units[unit])


def parse_size(size: str | int) -> int:
    if isinstance(size, int):
        return size
    return parse_human_readable_size(size)



from __future__ import annotations

import shlex
from dataclasses import dataclass
from getpass import getpass
from queue import Queue
from socket import timeout as timeout_error
from subprocess import PIPE, Popen
from typing import TYPE_CHECKING, Callable, Iterable, Optional, Union

import click
import gevent

from pyinfra import logger
from pyinfra.api import MaskString, QuoteString, StringCommand
from pyinfra.api.util import memoize

if TYPE_CHECKING:
    from pyinfra.api.arguments import ConnectorArguments
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


SUDO_ASKPASS_ENV_VAR = "PYINFRA_SUDO_PASSWORD"


SUDO_ASKPASS_COMMAND = r"""
temp=$(mktemp "${{TMPDIR:={0}}}/pyinfra-sudo-askpass-XXXXXXXXXXXX")
cat >"$temp"<<'__EOF__'
#!/bin/sh
printf '%s\n' "${1}"
__EOF__
chmod 755 "$temp"
echo "$temp"
"""


def run_local_process(
    command: str,
    stdin=None,
    timeout: Optional[int] = None,
    print_output: bool = False,
    print_prefix: str = "",
) -> tuple[int, "CommandOutput"]:
    process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE, stdin=PIPE)

    assert process.stdout is not None
    assert process.stderr is not None
    assert process.stdin is not None

    # Write any stdin and then close it
    if stdin:
        write_stdin(stdin, process.stdin)
    process.stdin.close()

    combined_output = read_output_buffers(
        process.stdout,
        process.stderr,
        timeout=timeout,
        print_output=print_output,
        print_prefix=print_prefix,
    )

    logger.debug("--> Waiting for exit status...")
    process.wait()
    logger.debug("--> Command exit status: %i", process.returncode)

    # Close any open file descriptors
    process.stdout.close()
    process.stderr.close()

    return process.returncode, combined_output


# Command output buffer handling
#


@dataclass
class OutputLine:
    buffer_name: str
    line: str


@dataclass
class CommandOutput:
    combined_lines: list[OutputLine]

    def __iter__(self):
        yield from self.combined_lines

    @property
    def output_lines(self) -> list[str]:
        return [line.line for line in self.combined_lines]

    @property
    def output(self) -> str:
        return "\n".join(self.output_lines)

    @property
    def stdout_lines(self) -> list[str]:
        return [line.line for line in self.combined_lines if line.buffer_name == "stdout"]

    @property
    def stdout(self) -> str:
        return "\n".join(self.stdout_lines)

    @property
    def stderr_lines(self) -> list[str]:
        return [line.line for line in self.combined_lines if line.buffer_name == "stderr"]

    @property
    def stderr(self) -> str:
        return "\n".join(self.stderr_lines)


def read_buffer(
    name: str,
    io: Iterable,
    output_queue: Queue[OutputLine],
    print_output=False,
    print_func=None,
) -> None:
    """
    Reads a file-like buffer object into lines and optionally prints the output.
    """

    def _print(line):
        if print_func:
            line = print_func(line)

        click.echo(line, err=True)

    for line in io:
        # Handle local Popen shells returning list of bytes, not strings
        if not isinstance(line, str):
            line = line.decode("utf-8")

        line = line.rstrip("\n")
        output_queue.put(OutputLine(name, line))

        if print_output:
            _print(line)


def read_output_buffers(
    stdout_buffer: Iterable,
    stderr_buffer: Iterable,
    timeout: Optional[int],
    print_output: bool,
    print_prefix: str,
) -> CommandOutput:
    output_queue: Queue[OutputLine] = Queue()

    # Iterate through outputs to get an exit status and generate desired list
    # output, done in two greenlets so stdout isn't printed before stderr. Not
    # attached to state.pool to avoid blocking it with 2x n-hosts greenlets.
    stdout_reader = gevent.spawn(
        read_buffer,
        "stdout",
        stdout_buffer,
        output_queue,
        print_output=print_output,
        print_func=lambda line: "{0}{1}".format(print_prefix, line),
    )
    stderr_reader = gevent.spawn(
        read_buffer,
        "stderr",
        stderr_buffer,
        output_queue,
        print_output=print_output,
        print_func=lambda line: "{0}{1}".format(
            print_prefix,
            click.style(line, "red"),
        ),
    )

    # Wait on output, with our timeout (or None)
    greenlets = gevent.wait((stdout_reader, stderr_reader), timeout=timeout)

    # Timeout doesn't raise an exception, but gevent.wait returns the greenlets
    # which did complete. So if both haven't completed, we kill them and fail
    # with a timeout.
    if len(greenlets) != 2:
        stdout_reader.kill()
        stderr_reader.kill()

        raise timeout_error()

    return CommandOutput(list(output_queue.queue))


# Connector execution control
#


def execute_command_with_sudo_retry(
    host: "Host",
    command_arguments: "ConnectorArguments",
    execute_command: Callable[..., tuple[int, CommandOutput]],
) -> tuple[int, CommandOutput]:
    return_code, output = execute_command()

    # If we failed look for a sudo password prompt line and re-submit using the sudo password. Look
    # at all lines here in case anything else gets printed, eg in:
    # https://github.com/pyinfra-dev/pyinfra/issues/1292
    if return_code != 0 and output and output.combined_lines:
        for line in reversed(output.combined_lines):
            if line.line.strip() == "sudo: a password is required":
                # If we need a password, ask the user for it and attach to the host
                # internal connector data for use when executing future commands.
                sudo_password = getpass("{0}sudo password: ".format(host.print_prefix))
                host.connector_data["prompted_sudo_password"] = sudo_password
                return_code, output = execute_command()
                break

    return return_code, output


def write_stdin(stdin, buffer):
    if hasattr(stdin, "readlines"):
        stdin = stdin.readlines()
    if not isinstance(stdin, (list, tuple)):
        stdin = [stdin]

    for line in stdin:
        if not line.endswith("\n"):
            line = "{0}\n".format(line)
        line = line.encode()
        buffer.write(line)
    buffer.close()


def remove_any_sudo_askpass_file(host) -> None:
    sudo_askpass_path = host.connector_data.get("sudo_askpass_path")
    if sudo_askpass_path:
        host.run_shell_command("rm -f {0}".format(sudo_askpass_path))
        host.connector_data["sudo_askpass_path"] = None


@memoize
def _show_use_su_login_warning() -> None:
    logger.warning(
        (
            "Using `use_su_login` may not work: "
            "some systems (MacOS, OpenBSD) ignore the flag when executing a command, "
            "use `sudo` + `use_sudo_login` instead."
        ),
    )


def extract_control_arguments(arguments: "ConnectorArguments") -> "ConnectorArguments":
    control_arguments: "ConnectorArguments" = {}

    if "_success_exit_codes" in arguments:
        control_arguments["_success_exit_codes"] = arguments.pop("_success_exit_codes")
    if "_timeout" in arguments:
        control_arguments["_timeout"] = arguments.pop("_timeout")
    if "_get_pty" in arguments:
        control_arguments["_get_pty"] = arguments.pop("_get_pty")
    if "_stdin" in arguments:
        control_arguments["_stdin"] = arguments.pop("_stdin")

    return control_arguments


def _ensure_sudo_askpass_set_for_host(host: "Host"):
    if host.connector_data.get("sudo_askpass_path"):
        return
    _, output = host.run_shell_command(
        SUDO_ASKPASS_COMMAND.format(host.get_temp_dir_config(), SUDO_ASKPASS_ENV_VAR)
    )
    host.connector_data["sudo_askpass_path"] = shlex.quote(output.stdout_lines[0])


def make_unix_command_for_host(
    state: "State",
    host: "Host",
    command: StringCommand,
    **command_arguments,
) -> StringCommand:
    if not command_arguments.get("_sudo"):
        # If no sudo, we've nothing to do here
        return make_unix_command(command, **command_arguments)

    # If the sudo password is not set in the direct arguments,
    # set it from the connector data value.
    if "_sudo_password" not in command_arguments or not command_arguments["_sudo_password"]:
        command_arguments["_sudo_password"] = host.connector_data.get("prompted_sudo_password")

    if command_arguments["_sudo_password"]:
        # Ensure the askpass path is correctly set and passed through
        _ensure_sudo_askpass_set_for_host(host)
        command_arguments["_sudo_askpass_path"] = host.connector_data["sudo_askpass_path"]
    return make_unix_command(command, **command_arguments)


# Connector command generation
#


def make_unix_command(
    command: StringCommand,
    _env=None,
    _chdir=None,
    _shell_executable="sh",
    # Su config
    _su_user=None,
    _use_su_login=False,
    _su_shell=None,
    _preserve_su_env=False,
    # Sudo config
    _sudo=False,
    _sudo_user=None,
    _use_sudo_login=False,
    _sudo_password="",
    _sudo_askpass_path=None,
    _preserve_sudo_env=False,
    # Doas config
    _doas=False,
    _doas_user=None,
    # Retry config (ignored in command generation but passed through)
    _retries=0,
    _retry_delay=0,
    _retry_until=None,
    # Temp dir config (ignored in command generation, used for temp file path generation)
    _temp_dir=None,
) -> StringCommand:
    """
    Builds a shell command with various kwargs.
    """

    if _shell_executable is not None and not isinstance(_shell_executable, str):
        _shell_executable = "sh"

    if _env:
        env_string = " ".join(['"{0}={1}"'.format(key, value) for key, value in _env.items()])
        command = StringCommand("export", env_string, "&&", command)

    if _chdir:
        command = StringCommand("cd", _chdir, "&&", command)

    command_bits: list[Union[str, StringCommand, QuoteString]] = []

    if _doas:
        command_bits.extend(["doas", "-n"])

        if _doas_user:
            command_bits.extend(["-u", _doas_user])

    if _sudo_password and _sudo_askpass_path:
        command_bits.extend(
            [
                "env",
                "SUDO_ASKPASS={0}".format(_sudo_askpass_path),
                MaskString("{0}={1}".format(SUDO_ASKPASS_ENV_VAR, shlex.quote(_sudo_password))),
            ],
        )

    if _sudo:
        command_bits.extend(["sudo", "-H"])

        if _sudo_password:
            command_bits.extend(["-A", "-k"])  # use askpass, disable cache
        else:
            command_bits.append("-n")  # disable prompt/interactivity

        if _use_sudo_login:
            command_bits.append("-i")

        if _preserve_sudo_env:
            command_bits.append("-E")

        if _sudo_user:
            command_bits.extend(("-u", _sudo_user))

    if _su_user:
        command_bits.append("su")

        if _use_su_login:
            _show_use_su_login_warning()
            command_bits.append("-l")

        if _preserve_su_env:
            command_bits.append("-m")

        if _su_shell:
            command_bits.extend(["-s", "`which {0}`".format(_su_shell)])

        command_bits.extend([_su_user, "-c"])

        if _shell_executable is not None:
            # Quote the whole shell -c 'command' as BSD `su` does not have a shell option
            command_bits.append(
                QuoteString(StringCommand(_shell_executable, "-c", QuoteString(command))),
            )
        else:
            command_bits.append(QuoteString(StringCommand(command)))
    else:
        if _shell_executable is not None:
            command_bits.extend([_shell_executable, "-c", QuoteString(command)])
        else:
            command_bits.extend([command])

    return StringCommand(*command_bits)


def make_win_command(command):
    """
    Builds a windows command with various kwargs.
    """

    # Quote the command as a string
    command = shlex.quote(str(command))
    command = "{0}".format(command)

    return command



"""
Manager Docker containers, volumes and networks. These operations allow you to manage Docker from
the view of the current inventory host. See the :doc:`../connectors/docker` to use Docker containers
as inventory directly.
"""

from __future__ import annotations

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.docker import (
    DockerContainer,
    DockerImage,
    DockerNetwork,
    DockerPlugin,
    DockerVolume,
)

from .util.docker import ContainerSpec, handle_docker, parse_image_reference


@operation()
def container(
    container: str,
    image: str = "",
    ports: list[str] | None = None,
    networks: list[str] | None = None,
    volumes: list[str] | None = None,
    env_vars: list[str] | None = None,
    labels: list[str] | None = None,
    pull_always: bool = False,
    present: bool = True,
    force: bool = False,
    start: bool = True,
    restart_policy: str | None = None,
    auto_remove: bool = False,
):
    """
    Manage Docker containers

    + container: name to identify the container
    + image: container image and tag ex: nginx:alpine
    + networks: network list to attach on container
    + ports: port list to expose
    + volumes: volume list to map on container
    + env_vars: environment variable list to inject on container
    + labels: Label list to attach to the container
    + pull_always: force image pull
    + force: remove a container with same name and create a new one
    + present: whether the container should be up and running
    + start: start or stop the container
    + restart_policy: restart policy to apply when a container exits
    + auto_remove: automatically remove the container and its associated anonymous volumes when it exits

    **Examples:**

    .. code:: python

        from pyinfra.operations import docker
        # Run a container
        docker.container(
            name="Deploy Nginx container",
            container="nginx",
            image="nginx:alpine",
            ports=["80:80"],
            present=True,
            force=True,
            networks=["proxy", "services"],
            volumes=["nginx_data:/usr/share/nginx/html"],
            pull_always=True,
            restart_policy="unless-stopped",
            auto_remove=True,
        )

        # Stop a container
        docker.container(
            name="Stop Nginx container",
            container="nginx",
            start=False,
        )

        # Start a container
        docker.container(
            name="Start Nginx container",
            container="nginx",
            start=True,
        )
    """

    want_spec = ContainerSpec(
        image,
        ports or list(),
        networks or list(),
        volumes or list(),
        env_vars or list(),
        labels or list(),
        pull_always,
        restart_policy,
        auto_remove,
    )
    existent_container = host.get_fact(DockerContainer, object_id=container)

    container_spec_changes = want_spec.diff_from_inspect(existent_container)

    is_running = (
        (existent_container[0]["State"]["Status"] == "running")
        if existent_container and existent_container[0]
        else False
    )
    recreating = existent_container and (force or container_spec_changes)
    removing = existent_container and not present

    do_remove = recreating or removing
    do_create = (present and not existent_container) or recreating
    do_start = start and (recreating or not is_running)
    do_stop = not start and not removing and is_running

    if do_remove:
        yield handle_docker(
            resource="container",
            command="remove",
            container=container,
        )

    if do_create:
        yield handle_docker(
            resource="container",
            command="create",
            container=container,
            spec=want_spec,
        )

    if do_start:
        yield handle_docker(
            resource="container",
            command="start",
            container=container,
        )

    if do_stop:
        yield handle_docker(
            resource="container",
            command="stop",
            container=container,
        )


@operation()
def image(image: str, present: bool = True, force: bool = False):
    """
    Manage Docker images

    + image: Image and tag ex: nginx:alpine
    + present: whether the Docker image should exist
    + force: always pull the image if present is True

    **Examples:**

    .. code:: python

        # Pull a Docker image
        docker.image(
            name="Pull nginx image",
            image="nginx:alpine",
            present=True,
        )

        # Remove a Docker image
        docker.image(
            name="Remove nginx image",
            image:"nginx:image",
            present=False,
        )
    """
    image_info = parse_image_reference(image)
    if present:
        if force:
            # always pull the image if force is True
            yield handle_docker(
                resource="image",
                command="pull",
                image=image,
            )
            return
        else:
            existent_image = host.get_fact(DockerImage, object_id=image)
            if image_info.digest:
                # If a digest is specified, we must ensure the exact image is present
                if existent_image:
                    host.noop(f"Image with digest {image_info.digest} already exists!")
                else:
                    yield handle_docker(
                        resource="image",
                        command="pull",
                        image=image,
                    )
            elif image_info.tag == "latest" or not image_info.tag:
                # If the tag is 'latest' or not specified, always pull to ensure freshness
                yield handle_docker(
                    resource="image",
                    command="pull",
                    image=image,
                )
            else:
                # For other tags, check if the image exists
                if existent_image:
                    host.noop(f"Image with tag {image_info.tag} already exists!")
                else:
                    yield handle_docker(
                        resource="image",
                        command="pull",
                        image=image,
                    )
    else:
        existent_image = host.get_fact(DockerImage, object_id=image)
        if existent_image:
            yield handle_docker(
                resource="image",
                command="remove",
                image=image,
            )
        else:
            host.noop("There is no {0} image!".format(image))


@operation()
def volume(volume: str, driver: str = "", labels: list[str] | None = None, present: bool = True):
    """
    Manage Docker volumes

    + volume: Volume name
    + driver: Docker volume storage driver
    + labels: Label list to attach in the volume
    + present: whether the Docker volume should exist

    **Examples:**

    .. code:: python

        # Create a Docker volume
        docker.volume(
            name="Create nginx volume",
            volume="nginx_data",
            present=True
        )
    """

    existent_volume = host.get_fact(DockerVolume, object_id=volume)

    if present:
        if existent_volume:
            host.noop("Volume already exists!")
            return

        yield handle_docker(
            resource="volume",
            command="create",
            volume=volume,
            driver=driver,
            labels=labels,
            present=present,
        )

    else:
        if existent_volume is None:
            host.noop("There is no {0} volume!".format(volume))
            return

        yield handle_docker(
            resource="volume",
            command="remove",
            volume=volume,
        )


@operation()
def network(
    network: str,
    driver: str = "",
    gateway: str = "",
    ip_range: str = "",
    ipam_driver: str = "",
    subnet: str = "",
    scope: str = "",
    aux_addresses: dict[str, str] | None = None,
    opts: list[str] | None = None,
    ipam_opts: list[str] | None = None,
    labels: list[str] | None = None,
    ingress: bool = False,
    attachable: bool = False,
    present: bool = True,
):
    """
    Manage docker networks

    + network: Network name
    + driver: Network driver ex: bridge or overlay
    + gateway: IPv4 or IPv6 Gateway for the master subnet
    + ip_range: Allocate container ip from a sub-range
    + ipam_driver: IP Address Management Driver
    + subnet: Subnet in CIDR format that represents a network segment
    + scope: Control the network's scope
    + aux_addresses: named aux addresses for the network
    + opts: Set driver specific options
    + ipam_opts: Set IPAM driver specific options
    + labels: Label list to attach in the network
    + ingress: Create swarm routing-mesh network
    + attachable: Enable manual container attachment
    + present: whether the Docker network should exist

    **Examples:**

    .. code:: python

        # Create Docker network
        docker.network(
            network="nginx",
            attachable=True,
            present=True,
        )
    """
    existent_network = host.get_fact(DockerNetwork, object_id=network)

    if present:
        if existent_network:
            host.noop("Network {0} already exists!".format(network))
            return

        yield handle_docker(
            resource="network",
            command="create",
            network=network,
            driver=driver,
            gateway=gateway,
            ip_range=ip_range,
            ipam_driver=ipam_driver,
            subnet=subnet,
            scope=scope,
            aux_addresses=aux_addresses,
            opts=opts,
            ipam_opts=ipam_opts,
            labels=labels,
            ingress=ingress,
            attachable=attachable,
            present=present,
        )

    else:
        if existent_network is None:
            host.noop("Network {0} does not exist!".format(network))
            return

        yield handle_docker(
            resource="network",
            command="remove",
            network=network,
        )


@operation(is_idempotent=False)
def prune(
    all: bool = False,
    volumes: bool = False,
    filter: str = "",
):
    """
    Execute a docker system prune.

    + all: Remove all unused images not just dangling ones
    + volumes: Prune anonymous volumes
    + filter: Provide filter values (e.g. "label=<key>=<value>" or "until=24h")

    **Examples:**

    .. code:: python

        # Remove dangling images
        docker.prune(
            name="remove dangling images",
        )

        # Remove all images and volumes
        docker.prune(
            name="Remove all images and volumes",
            all=True,
            volumes=True,
        )

        # Remove images older than 90 days
        docker.prune(
            name="Remove unused older than 90 days",
            filter="until=2160h"
        )
    """

    yield handle_docker(
        resource="system",
        command="prune",
        all=all,
        volumes=volumes,
        filter=filter,
    )


@operation()
def plugin(
    plugin: str,
    alias: str | None = None,
    present: bool = True,
    enabled: bool = True,
    plugin_options: dict[str, str] | None = None,
):
    """
    Manage Docker plugins

    + plugin: Plugin name
    + alias: Alias for the plugin (optional)
    + present: Whether the plugin should be installed
    + enabled: Whether the plugin should be enabled
    + plugin_options: Options to pass to the plugin

    **Examples:**

    .. code:: python

        # Install and enable a Docker plugin
        docker.plugin(
            name="Install and enable a Docker plugin",
            plugin="username/my-awesome-plugin:latest",
            alias="my-plugin",
            present=True,
            enabled=True,
            plugin_options={"option1": "value1", "option2": "value2"},
        )
    """
    plugin_name = alias if alias else plugin
    existent_plugin = host.get_fact(DockerPlugin, object_id=plugin_name)
    if existent_plugin:
        existent_plugin = existent_plugin[0]

    if present:
        if existent_plugin:
            plugin_options_different = (
                plugin_options and existent_plugin["Settings"]["Env"] != plugin_options
            )
            if plugin_options_different:
                # Update options on existing plugin
                if existent_plugin["Enabled"]:
                    yield handle_docker(
                        resource="plugin",
                        command="disable",
                        plugin=plugin_name,
                    )
                yield handle_docker(
                    resource="plugin",
                    command="set",
                    plugin=plugin_name,
                    enabled=enabled,
                    existent_options=existent_plugin["Settings"]["Env"],
                    required_options=plugin_options,
                )
                if enabled:
                    yield handle_docker(
                        resource="plugin",
                        command="enable",
                        plugin=plugin_name,
                    )
            else:
                # Options are the same, check if enabled state is different
                if existent_plugin["Enabled"] == enabled:
                    host.noop(
                        f"Plugin '{plugin_name}' is already installed with the same options "
                        f"and {'enabled' if enabled else 'disabled'}."
                    )
                    return
                else:
                    command = "enable" if enabled else "disable"
                    yield handle_docker(
                        resource="plugin",
                        command=command,
                        plugin=plugin_name,
                    )
        else:
            yield handle_docker(
                resource="plugin",
                command="install",
                plugin=plugin,
                alias=alias,
                enabled=enabled,
                plugin_options=plugin_options,
            )
    else:
        if not existent_plugin:
            host.noop(f"Plugin '{plugin_name}' is not installed.")
            return
        yield handle_docker(
            resource="plugin",
            command="remove",
            plugin=plugin_name,
        )



import abc
import sys
from inspect import getframeinfo
from traceback import format_exception, format_tb, walk_tb
from types import TracebackType

import click
from typing_extensions import override

from pyinfra import logger
from pyinfra.api.exceptions import (
    ArgumentTypeError,
    ConnectorDataTypeError,
    OperationError,
    PyinfraError,
)
from pyinfra.api.util import PYINFRA_INSTALL_DIR


def get_frame_line_from_tb(tb: TracebackType):
    frame_lines = list(walk_tb(tb))
    frame_lines.reverse()
    for frame, line in frame_lines:
        info = getframeinfo(frame)
        if info.filename.startswith(PYINFRA_INSTALL_DIR):
            continue
        return info


class WrappedError(click.ClickException):
    def __init__(self, e: Exception):
        self.traceback = e.__traceback__
        self.exception = e

        # Pull message from the wrapped exception
        message = getattr(e, "message", e.args[0])
        if not isinstance(message, str):
            message = repr(message)
        self.message = message

    @override
    def show(self, file=None):
        name = "unknown error"

        if isinstance(self.exception, ConnectorDataTypeError):
            name = "Connector data type error"
        elif isinstance(self.exception, ArgumentTypeError):
            name = "Argument type error"
        elif isinstance(self.exception, OperationError):
            name = "Operation error"
        elif isinstance(self.exception, PyinfraError):
            name = "pyinfra error"
        elif isinstance(self.exception, IOError):
            name = "Local IO error"

        if self.traceback:
            info = get_frame_line_from_tb(self.traceback)
            if info:
                name = f"{name} in {info.filename} line {info.lineno}"

        logger.warning(
            "--> {0}: {1}".format(
                click.style(name, "red", bold=True),
                self,
            ),
        )


class CliError(click.ClickException):
    @override
    def show(self, file=None):
        logger.warning(
            "--> {0}: {1}".format(
                click.style("pyinfra error", "red", bold=True),
                self,
            ),
        )


class UnexpectedMixin(abc.ABC):
    exception: Exception
    traceback: TracebackType

    def get_traceback_lines(self):
        traceback = getattr(self.exception, "_traceback")
        return format_tb(traceback)

    def get_traceback(self):
        return "".join(self.get_traceback_lines())

    def get_exception(self):
        return "".join(format_exception(self.exception.__class__, self.exception, None))


class UnexpectedExternalError(click.ClickException, UnexpectedMixin):
    def __init__(self, e, filename):
        _, _, traceback = sys.exc_info()
        e._traceback = traceback
        self.exception = e
        self.filename = filename

    @override
    def show(self, file=None):
        logger.warning(
            "--> {0}:\n".format(
                click.style(
                    "An exception occurred in: {0}".format(self.filename),
                    "red",
                    bold=True,
                ),
            ),
        )

        click.echo("Traceback (most recent call last):", err=True)
        click.echo(self.get_traceback(), err=True, nl=False)
        click.echo(self.get_exception(), err=True)


class UnexpectedInternalError(click.ClickException, UnexpectedMixin):
    def __init__(self, e):
        _, _, traceback = sys.exc_info()
        e._traceback = traceback
        self.exception = e

    @override
    def show(self, file=None):
        click.echo(
            "--> {0}:\n".format(
                click.style(
                    "An internal exception occurred",
                    "red",
                    bold=True,
                ),
            ),
            err=True,
        )

        traceback_lines = self.get_traceback_lines()
        traceback = self.get_traceback()

        # Syntax errors contain the filename/line/etc, but other exceptions
        # don't, so print the *last* call to stderr.
        if not isinstance(self.exception, SyntaxError):
            sys.stderr.write(traceback_lines[-1])

        exception = self.get_exception()
        click.echo(exception, err=True)

        with open("pyinfra-debug.log", "w", encoding="utf-8") as f:
            f.write(traceback)
            f.write(exception)

        logger.debug(traceback)
        logger.debug(exception)

        click.echo(
            "--> The full traceback has been written to {0}".format(
                click.style("pyinfra-debug.log", bold=True),
            ),
            err=True,
        )
        click.echo(
            (
                "--> If this is unexpected please consider submitting a bug report "
                "on GitHub, for more information run `pyinfra --support`."
            ),
            err=True,
        )



from pyinfra.api import Config, State, StringCommand
from pyinfra.api.connect import connect_all, disconnect_all
from pyinfra.api.deploy import add_deploy, deploy
from pyinfra.api.operations import run_ops
from pyinfra.api.state import StateStage
from pyinfra.operations import server

from ..paramiko_util import PatchSSHTestCase
from ..util import make_inventory


class TestDeploysApi(PatchSSHTestCase):
    def test_deploy(self):
        inventory = make_inventory()
        somehost = inventory.get_host("somehost")
        anotherhost = inventory.get_host("anotherhost")

        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare

        # Enable printing on this test to catch any exceptions in the formatting
        state.print_output = True
        state.print_input = True
        state.print_fact_info = True
        state.print_noop_info = True

        connect_all(state)

        @deploy()
        def test_deploy(state=None, host=None):
            server.shell(commands=["echo first command"])
            server.shell(commands=["echo second command"])

        add_deploy(state, test_deploy)

        op_order = state.get_op_order()

        # Ensure we have an op
        assert len(op_order) == 2

        # Ensure run ops works
        run_ops(state)

        first_op_hash = op_order[0]
        assert state.op_meta[first_op_hash].names == {"test_deploy | server.shell"}
        assert state.ops[somehost][first_op_hash].operation_meta._commands == [
            StringCommand("echo first command"),
        ]
        assert state.ops[anotherhost][first_op_hash].operation_meta._commands == [
            StringCommand("echo first command"),
        ]

        second_op_hash = op_order[1]
        assert state.op_meta[second_op_hash].names == {"test_deploy | server.shell"}
        assert state.ops[somehost][second_op_hash].operation_meta._commands == [
            StringCommand("echo second command"),
        ]
        assert state.ops[anotherhost][second_op_hash].operation_meta._commands == [
            StringCommand("echo second command"),
        ]

        # Ensure ops completed OK
        assert state.results[somehost].success_ops == 2
        assert state.results[somehost].ops == 2
        assert state.results[anotherhost].success_ops == 2
        assert state.results[anotherhost].ops == 2

        # And w/o errors
        assert state.results[somehost].error_ops == 0
        assert state.results[anotherhost].error_ops == 0

        disconnect_all(state)

    def test_nested_deploy(self):
        inventory = make_inventory()
        somehost = inventory.get_host("somehost")

        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare

        # Enable printing on this test to catch any exceptions in the formatting
        state.print_output = True
        state.print_input = True
        state.print_fact_info = True
        state.print_noop_info = True

        connect_all(state)

        @deploy()
        def test_nested_deploy():
            server.shell(commands=["echo nested command"])

        @deploy()
        def test_deploy():
            server.shell(commands=["echo first command"])
            test_nested_deploy()
            server.shell(commands=["echo second command"])

        add_deploy(state, test_deploy)

        op_order = state.get_op_order()

        # Ensure we have an op
        assert len(op_order) == 3

        # Ensure run ops works
        run_ops(state)

        first_op_hash = op_order[0]
        assert state.op_meta[first_op_hash].names == {"test_deploy | server.shell"}
        assert state.ops[somehost][first_op_hash].operation_meta._commands == [
            StringCommand("echo first command"),
        ]

        second_op_hash = op_order[1]
        assert state.op_meta[second_op_hash].names == {
            "test_deploy | test_nested_deploy | server.shell",
        }
        assert state.ops[somehost][second_op_hash].operation_meta._commands == [
            StringCommand("echo nested command"),
        ]

        third_op_hash = op_order[2]
        assert state.op_meta[third_op_hash].names == {"test_deploy | server.shell"}
        assert state.ops[somehost][third_op_hash].operation_meta._commands == [
            StringCommand("echo second command"),
        ]
